@@ -79,6 +79,7 @@ class MemcachedLock(locking.Lock):
 
 
 class MemcachedDriver(coordination.CoordinationDriver):
+    """A memcached based driver."""
 
     _GROUP_PREFIX = b'_TOOZ_GROUP_'
     _GROUP_LEADER_PREFIX = b'_TOOZ_GROUP_LEADER_'
@@ -175,6 +176,20 @@ class MemcachedDriver(coordination.CoordinationDriver):
                 # Someone updated the group list before us, try again!
                 raise _retry.Retry
 
+    @_retry.retry
+    def _remove_from_group_list(self, group_id):
+        """Remove group from the group list.
+
+        :param group_id: The group id
+        """
+        group_list, cas = self.client.gets(self._GROUP_LIST_KEY)
+        group_list = set(group_list)
+        group_list.remove(group_id)
+        if not self.client.cas(self._GROUP_LIST_KEY,
+                               list(group_list), cas):
+            # Someone updated the group list before us, try again!
+            raise _retry.Retry
+
     def create_group(self, group_id):
         encoded_group = self._encode_group_id(group_id)
 
@@ -196,7 +211,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
         @_retry.retry
         def _join_group():
             group_members, cas = self.client.gets(encoded_group)
-            if not cas:
+            if group_members is None:
                 raise coordination.GroupNotCreated(group_id)
             if self._member_id in group_members:
                 raise coordination.MemberAlreadyExist(group_id,
@@ -217,7 +232,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
         @_retry.retry
         def _leave_group():
             group_members, cas = self.client.gets(encoded_group)
-            if not cas:
+            if group_members is None:
                 raise coordination.GroupNotCreated(group_id)
             if self._member_id not in group_members:
                 raise coordination.MemberNotJoined(group_id, self._member_id)
@@ -229,12 +244,46 @@ class MemcachedDriver(coordination.CoordinationDriver):
 
         return MemcachedFutureResult(self._executor.submit(_leave_group))
 
+    def _destroy_group(self, group_id):
+        self.client.delete(self._encode_group_id(group_id))
+
+    def delete_group(self, group_id):
+        encoded_group = self._encode_group_id(group_id)
+
+        @_retry.retry
+        def _delete_group():
+            group_members, cas = self.client.gets(encoded_group)
+            if group_members is None:
+                raise coordination.GroupNotCreated(group_id)
+            if group_members != {}:
+                raise coordination.GroupNotEmpty(group_id)
+            # Delete is not atomic, so we first set the group to
+            # using CAS, and then we delete it, to avoid race conditions.
+            if not self.client.cas(encoded_group, None, cas):
+                raise _retry.Retry
+            self.client.delete(encoded_group)
+            self._remove_from_group_list(group_id)
+
+        return MemcachedFutureResult(self._executor.submit(_delete_group))
+
+    @_retry.retry
     def _get_members(self, group_id):
-        group_members = self.client.get(self._encode_group_id(group_id))
+        encoded_group = self._encode_group_id(group_id)
+        group_members, cas = self.client.gets(encoded_group)
         if group_members is None:
             raise coordination.GroupNotCreated(group_id)
-        return dict((m, v) for m, v in six.iteritems(group_members)
-                    if self.client.get(self._encode_member_id(m)))
+        actual_group_members = {}
+        for m, v in six.iteritems(group_members):
+            # Never kick self from the group, we know we're alive
+            if (m == self._member_id
+               or self.client.get(self._encode_member_id(m))):
+                actual_group_members[m] = v
+        if group_members != actual_group_members:
+            # There are some dead members, update the group
+            if not self.client.cas(encoded_group, actual_group_members, cas):
+                # It changed, let's try again
+                raise _retry.Retry
+        return actual_group_members
 
     def get_members(self, group_id):
         def _get_members():
@@ -256,7 +305,7 @@ class MemcachedDriver(coordination.CoordinationDriver):
         @_retry.retry
         def _update_capabilities():
             group_members, cas = self.client.gets(encoded_group)
-            if cas is None:
+            if group_members is None:
                 raise coordination.GroupNotCreated(group_id)
             if self._member_id not in group_members:
                 raise coordination.MemberNotJoined(group_id, self._member_id)
@@ -325,7 +374,10 @@ class MemcachedDriver(coordination.CoordinationDriver):
     def run_watchers(self):
         result = []
         for group_id in self.client.get(self._GROUP_LIST_KEY):
-            group_members = set(self._get_members(group_id))
+            try:
+                group_members = set(self._get_members(group_id))
+            except coordination.GroupNotCreated:
+                group_members = set()
             old_group_members = self._group_members[group_id]
 
             for member_id in (group_members - old_group_members):
