@@ -17,6 +17,7 @@
 from __future__ import absolute_import
 
 import contextlib
+from distutils import version
 import logging
 
 from concurrent import futures
@@ -24,7 +25,11 @@ import msgpack
 from oslo.utils import strutils
 import redis
 from redis import exceptions
+from redis import lock as redis_locks
+from redis import sentinel
 import six
+from six.moves import map as compat_map
+from six.moves import zip as compat_zip
 
 import tooz
 from tooz import coordination
@@ -34,6 +39,36 @@ from tooz import utils
 LOG = logging.getLogger(__name__)
 
 
+def _version_checker(version_required):
+    """Checks server version is supported *before* running decorated method."""
+
+    if isinstance(version_required, six.string_types):
+        desired_version = version.LooseVersion(version_required)
+    elif isinstance(version_required, version.LooseVersion):
+        desired_version = version_required
+    else:
+        raise TypeError("Version decorator expects a string/version type")
+
+    def wrapper(meth):
+
+        @six.wraps(meth)
+        def decorator(self, *args, **kwargs):
+            useable, redis_version = self._check_fetch_redis_version(
+                desired_version)
+            if not useable:
+                raise tooz.NotImplemented("Redis version greater than or"
+                                          " equal to '%s' is required"
+                                          " to use this feature; '%s' is"
+                                          " being used which is not new"
+                                          " enough" % (desired_version,
+                                                       redis_version))
+            return meth(self, *args, **kwargs)
+
+        return decorator
+
+    return wrapper
+
+
 @contextlib.contextmanager
 def _translate_failures():
     """Translates common redis exceptions into tooz exceptions."""
@@ -48,7 +83,15 @@ def _translate_failures():
 class RedisLock(locking.Lock):
     def __init__(self, coord, client, name, timeout):
         self._name = "%s_%s_lock" % (coord.namespace, six.text_type(name))
-        self._lock = client.lock(self._name, timeout=timeout)
+        # Avoid using lua locks to keep compatible with more versions
+        # of redis (and not just the ones that have lua support, also avoids
+        # ones that don't appear to have fully working lua support...)
+        #
+        # Issue opened: https://github.com/andymccurdy/redis-py/issues/550
+        #
+        # When that gets fixed (and detects the servers capabilities better
+        # we can likely turn this back on to being smart).
+        self._lock = redis_locks.Lock(client, self._name, timeout=timeout)
         self._coord = coord
         self._acquired = False
 
@@ -101,19 +144,49 @@ class RedisDriver(coordination.CoordinationDriver):
       some notion of HA (values *can* be lost when a failover transition
       occurs).
 
+    To use a sentinel the connection URI must point to the Sentinel server.
+    At connection time the sentinel will be asked for the current IP and port
+    of the master and then connect there. The connection URI for sentinel
+    should be written as follows::
+
+      redis://<sentinel host>:<sentinel port>?sentinel=<master name>
+
     Further resources/links:
 
     - http://redis.io/
     - http://redis.io/topics/sentinel
     - http://redis.io/topics/cluster-spec
+
+    Note that this client will itself retry on transaction failure (when they
+    keys being watched have changed underneath the current transaction).
+    Currently the number of attempts that are tried is infinite (this might
+    be addressed in https://github.com/andymccurdy/redis-py/issues/566 when
+    that gets worked on). See http://redis.io/topics/transactions for more
+    information on this topic.
     """
 
+    # The min redis version that this driver requires to operate with...
+    #
+    # NOTE(harlowja): 2.2.0 was selected since its the current version that
+    # exists on ubuntu precise, that version will work to a degree (except
+    # locks will not work with that version), in the future we can raise this
+    # as we move off of precise (and/or need newer features that the older
+    # versions of redis just don't have).
+    _MIN_VERSION = version.LooseVersion("2.2.0")
+
     # Redis deletes dictionaries that have no keys in them, which means the
     # key will disappear which means we can't tell the difference between
     # a group not existing and a group being empty without this key being
     # saved...
-    _GROUP_EXISTS = '__created__'
-    _NAMESPACE_SEP = ':'
+    _GROUP_EXISTS = b'__created__'
+    _NAMESPACE_SEP = b':'
+
+    # This is for python3.x; which will behave differently when returned
+    # binary types or unicode types (redis uses binary internally it appears),
+    # so to just stick with a common way of doing this, make all the things
+    # binary (with this default encoding if one is not given and a unicode
+    # string is provided).
+    _DEFAULT_ENCODING = 'utf8'
 
     # These are used when extracting options from to make a client.
     #
@@ -128,6 +201,7 @@ class RedisDriver(coordination.CoordinationDriver):
         'ssl',
         'ssl_certfile',
         'ssl_keyfile',
+        'sentinel',
     ])
     _CLIENT_BOOL_ARGS = frozenset([
         'retry_on_timeout',
@@ -144,20 +218,50 @@ class RedisDriver(coordination.CoordinationDriver):
         super(RedisDriver, self).__init__()
         self._parsed_url = parsed_url
         self._options = options
+        encoding = options.get('encoding', [self._DEFAULT_ENCODING])
+        self._encoding = encoding[-1]
         timeout = options.get('timeout', [self._CLIENT_DEFAULT_SOCKET_TO])
         self.timeout = int(timeout[-1])
+        self.membership_timeout = float(options.get(
+            'membership_timeout', timeout)[-1])
         lock_timeout = options.get('lock_timeout', [self.timeout])
         self.lock_timeout = int(lock_timeout[-1])
-        self._namespace = options.get('namespace', '_tooz')
-        self._group_prefix = "%s_group" % (self._namespace)
-        self._leader_prefix = "%s_leader" % (self._namespace)
-        self._groups = "%s_groups" % (self._namespace)
+        namespace = options.get('namespace', ['_tooz'])[-1]
+        self._namespace = self._to_binary(namespace)
+        self._group_prefix = self._namespace + b"_group"
+        self._leader_prefix = self._namespace + b"_leader"
+        self._beat_prefix = self._namespace + b"_beats"
+        self._groups = self._namespace + b"_groups"
         self._client = None
-        self._member_id = member_id
+        self._member_id = self._to_binary(member_id)
         self._acquired_locks = set()
         self._joined_groups = set()
         self._executor = None
         self._started = False
+        self._server_info = {}
+
+    def _check_fetch_redis_version(self, geq_version, not_existent=True):
+        if isinstance(geq_version, six.string_types):
+            desired_version = version.LooseVersion(geq_version)
+        elif isinstance(geq_version, version.LooseVersion):
+            desired_version = geq_version
+        else:
+            raise TypeError("Version check expects a string/version type")
+        try:
+            redis_version = version.LooseVersion(
+                    self._server_info['redis_version'])
+        except KeyError:
+            return (not_existent, None)
+        else:
+            if redis_version < desired_version:
+                return (False, redis_version)
+            else:
+                return (True, redis_version)
+
+    def _to_binary(self, text):
+        if not isinstance(text, six.binary_type):
+            text = text.encode(self._encoding)
+        return text
 
     @property
     def namespace(self):
@@ -167,6 +271,9 @@ class RedisDriver(coordination.CoordinationDriver):
     def running(self):
         return self._started
 
+    # 2.6.0 is required since internally PEXPIRE is used and that was only
+    # added in 2.6.0; so avoid using this on versions less than that...
+    @_version_checker("2.6.0")
     def get_lock(self, name):
         return RedisLock(self, self._client, name, self.lock_timeout)
 
@@ -217,6 +324,18 @@ class RedisDriver(coordination.CoordinationDriver):
             kwargs[a] = v
         if 'socket_timeout' not in kwargs:
             kwargs['socket_timeout'] = default_socket_timeout
+
+        # Ask the sentinel for the current master if there is a
+        # sentinel arg.
+        if 'sentinel' in kwargs:
+            sentinel_server = sentinel.Sentinel(
+                [(kwargs['host'], kwargs['port'])],
+                socket_timeout=kwargs['socket_timeout'])
+            master_host, master_port = sentinel_server.discover_master(
+                kwargs['sentinel'])
+            kwargs['host'] = master_host
+            kwargs['port'] = master_port
+            del kwargs['sentinel']
         return redis.StrictRedis(**kwargs)
 
     def _start(self):
@@ -230,30 +349,61 @@ class RedisDriver(coordination.CoordinationDriver):
             # Ensure that the server is alive and not dead, this does not
             # ensure the server will always be alive, but does insure that it
             # at least is alive once...
+            with _translate_failures():
+                self._server_info = self._client.info()
+            # Validate we have a good enough redis version we are connected
+            # to so that the basic set of features we support will actually
+            # work (instead of blowing up).
+            new_enough, redis_version = self._check_fetch_redis_version(
+                self._MIN_VERSION)
+            if not new_enough:
+                raise tooz.NotImplemented("Redis version greater than or"
+                                          " equal to '%s' is required"
+                                          " to use this driver; '%s' is"
+                                          " being used which is not new"
+                                          " enough" % (self._MIN_VERSION,
+                                                       redis_version))
             self.heartbeat()
             self._started = True
 
-    @classmethod
-    def _encode_member_id(cls, member_id):
-        if member_id == cls._GROUP_EXISTS:
-            raise ValueError("Not allowed to use private keys as a member id")
-        return six.text_type(member_id)
+    def _encode_beat_id(self, member_id):
+        return self._NAMESPACE_SEP.join([self._beat_prefix,
+                                         self._to_binary(member_id)])
 
-    @staticmethod
-    def _decode_member_id(member_id):
+    def _encode_member_id(self, member_id):
+        member_id = self._to_binary(member_id)
+        if member_id == self._GROUP_EXISTS:
+            raise ValueError("Not allowed to use private keys as a member id")
         return member_id
 
-    def _encode_group_id(self, group_id):
-        return self._NAMESPACE_SEP.join([self._group_prefix,
-                                         six.text_type(group_id)])
+    def _decode_member_id(self, member_id):
+        return self._to_binary(member_id)
 
-    def _encode_group_leader(self, group_id):
-        return self._NAMESPACE_SEP.join([self._leader_prefix,
-                                         six.text_type(group_id)])
+    def _encode_group_id(self, group_id, apply_namespace=True):
+        group_id = self._to_binary(group_id)
+        if not apply_namespace:
+            return group_id
+        return self._NAMESPACE_SEP.join([self._group_prefix, group_id])
+
+    def _decode_group_id(self, group_id):
+        return self._to_binary(group_id)
 
     def heartbeat(self):
         with _translate_failures():
-            self._client.ping()
+            beat_id = self._encode_beat_id(self._member_id)
+            # Use milliseconds if we can (which are more accurate than
+            # just seconds); but this PSETEX support was added in 2.6.0 or
+            # newer so we can only use it then...
+            supports_psetex, _version = self._check_fetch_redis_version(
+                '2.6.0', not_existent=False)
+            if not supports_psetex:
+                expiry_secs = max(0, int(self.membership_timeout))
+                self._client.setex(beat_id, time=expiry_secs,
+                                   value=b"Not dead!")
+            else:
+                expiry_ms = max(0, int(self.membership_timeout * 1000.0))
+                self._client.psetex(beat_id, time_ms=expiry_ms,
+                                    value=b"Not dead!")
         for lock in self._acquired_locks:
             try:
                 lock.heartbeat()
@@ -272,6 +422,9 @@ class RedisDriver(coordination.CoordinationDriver):
             group_id = self._joined_groups.pop()
             try:
                 self.leave_group(group_id).get()
+            except (coordination.MemberNotJoined,
+                    coordination.GroupNotCreated):
+                pass
             except coordination.ToozError:
                 LOG.warning("Unable to leave group '%s'", group_id,
                             exc_info=True)
@@ -279,7 +432,18 @@ class RedisDriver(coordination.CoordinationDriver):
             self._executor.shutdown(wait=True)
             self._executor = None
         if self._client is not None:
+            # Make sure we no longer exist...
+            beat_id = self._encode_beat_id(self._member_id)
+            try:
+                # NOTE(harlowja): this will delete nothing if the key doesn't
+                # exist in the first place, which is fine/expected/desired...
+                with _translate_failures():
+                    self._client.delete(beat_id)
+            except coordination.ToozError:
+                LOG.warning("Unable to delete heartbeat key '%s'", beat_id,
+                            exc_info=True)
             self._client = None
+        self._server_info = {}
         self._started = False
 
     def _submit(self, cb, *args, **kwargs):
@@ -297,7 +461,8 @@ class RedisDriver(coordination.CoordinationDriver):
         def _create_group(p):
             if p.exists(encoded_group):
                 raise coordination.GroupAlreadyExist(group_id)
-            p.sadd(self._groups, group_id)
+            p.sadd(self._groups,
+                   self._encode_group_id(group_id, apply_namespace=False))
             # Add our special key to avoid redis from deleting the dictionary
             # when it becomes empty (which is not what we currently want)...
             p.hset(encoded_group, self._GROUP_EXISTS, '1')
@@ -348,11 +513,40 @@ class RedisDriver(coordination.CoordinationDriver):
         def _get_members(p):
             if not p.exists(encoded_group):
                 raise coordination.GroupNotCreated(group_id)
-            members = []
+            potential_members = []
             for m in p.hkeys(encoded_group):
+                m = self._decode_member_id(m)
                 if m != self._GROUP_EXISTS:
-                    members.append(self._decode_member_id(m))
-            return members
+                    potential_members.append(m)
+            if not potential_members:
+                return []
+            # Ok now we need to see which members have passed away...
+            gone_members = set()
+            member_values = p.mget(compat_map(self._encode_beat_id,
+                                              potential_members))
+            for (potential_member, value) in compat_zip(potential_members,
+                                                        member_values):
+                # Always preserve self (just incase we haven't heartbeated
+                # while this call/s was being made...), this does *not* prevent
+                # another client from removing this though...
+                if potential_member == self._member_id:
+                    continue
+                if not value:
+                    gone_members.add(potential_member)
+            # Trash all the members that no longer are with us... RIP...
+            if gone_members:
+                many_at_once, _version = self._check_fetch_redis_version(
+                    "2.4.0", not_existent=False)
+                if not many_at_once:
+                    for m in gone_members:
+                        p.hdel(encoded_group, self._encode_member_id(m))
+                else:
+                    encoded_gone_members = [self._encode_member_id(m)
+                                            for m in gone_members]
+                    p.hdel(encoded_group, *encoded_gone_members)
+                return [m for m in potential_members if m not in gone_members]
+            else:
+                return potential_members
 
         return RedisFutureResult(self._submit(self._client.transaction,
                                               _get_members, encoded_group,
@@ -396,12 +590,35 @@ class RedisDriver(coordination.CoordinationDriver):
                                               encoded_group,
                                               value_from_callable=True))
 
+    def delete_group(self, group_id):
+        encoded_group = self._encode_group_id(group_id)
+
+        def _delete_group(p):
+            # An empty group still have the special key _GROUP_EXISTS set, so
+            # its len is 1
+            if p.hlen(encoded_group) > 1:
+                raise coordination.GroupNotEmpty(group_id)
+            if not p.delete(encoded_group):
+                raise coordination.GroupNotCreated(group_id)
+            p.srem(self._groups,
+                   self._encode_group_id(group_id,
+                                         apply_namespace=False))
+
+        return RedisFutureResult(self._submit(self._client.transaction,
+                                              _delete_group,
+                                              encoded_group,
+                                              value_from_callable=True))
+
+    def _destroy_group(self, group_id):
+        """Should only be used in tests..."""
+        self._client.delete(self._encode_group_id(group_id))
+
     def get_groups(self):
 
         def _get_groups():
             results = []
             for g in self._client.smembers(self._groups):
-                results.append(g)
+                results.append(self._decode_group_id(g))
             return results
 
         return RedisFutureResult(self._submit(_get_groups))
@@ -435,8 +652,21 @@ class RedisDriver(coordination.CoordinationDriver):
     def run_watchers(self, timeout=None):
         result = []
         for group_id in self.get_groups().get(timeout=timeout):
-            group_members = set(self.get_members(group_id)
-                                .get(timeout=timeout))
+            try:
+                group_members = self.get_members(group_id).get(timeout=timeout)
+            except coordination.GroupNotCreated:
+                group_members = set()
+            else:
+                group_members = set(group_members)
+            # I was booted out...
+            #
+            # TODO(harlowja): perhaps we should have a way to notify
+            # watchers that this has happened (the below mechanism will
+            # also do this, but it might be better to have a separate
+            # way when 'self' membership is lost)?
+            if (group_id in self._joined_groups and
+                    self._member_id not in group_members):
+                self._joined_groups.discard(group_id)
             old_group_members = self._group_members.get(group_id, set())
             for member_id in (group_members - old_group_members):
                 result.extend(
