@@ -29,27 +29,22 @@ class TestAPI(testscenarios.TestWithScenarios,
               tests.TestCaseSkipNotImplemented):
 
     scenarios = [
-        ('zookeeper', {'url': 'kazoo://127.0.0.1:2181?timeout=5',
-                       'bad_url': 'kazoo://localhost:1'}),
+        ('kazoo', {'url': os.getenv("TOOZ_TEST_ZOOKEEPER_URL"),
+                   'bad_url': 'kazoo://localhost:1'}),
         ('zake', {'url': 'zake://?timeout=5'}),
-        ('memcached', {'url': 'memcached://?timeout=5',
-                       'bad_url': 'memcached://localhost:1'}),
+        ('memcached', {'url': os.getenv("TOOZ_TEST_MEMCACHED_URL"),
+                       'bad_url': 'memcached://localhost:1',
+                       'timeout_capable': True}),
         ('ipc', {'url': 'ipc://'}),
-        ('redis', {'url': 'redis://localhost:6379?timeout=5',
-                   'bad_url': 'redis://localhost:1'}),
+        ('redis', {'url': os.getenv("TOOZ_TEST_REDIS_URL"),
+                   'bad_url': 'redis://localhost:1',
+                   'timeout_capable': True}),
         ('postgresql', {'url': os.getenv("TOOZ_TEST_PGSQL_URL"),
                         'bad_url': 'postgresql://localhost:1'}),
         ('mysql', {'url': os.getenv("TOOZ_TEST_MYSQL_URL"),
                    'bad_url': 'mysql://localhost:1'}),
     ]
 
-    # Only certain drivers have the tested support for timeouts that we test
-    # here, these are the lists of driver types that do support our test type.
-    timeout_capable = [
-        'memcached://',
-        'redis://',
-    ]
-
     def assertRaisesAny(self, exc_classes, callable_obj, *args, **kwargs):
         checkers = [matchers.MatchesException(exc_class)
                     for exc_class in exc_classes]
@@ -57,16 +52,6 @@ class TestAPI(testscenarios.TestWithScenarios,
         callable_obj = testcase.Nullary(callable_obj, *args, **kwargs)
         self.assertThat(callable_obj, matcher)
 
-    def skipIfNotSupported(self, supported):
-        applicable = False
-        for prefix in supported:
-            if self.url.startswith(prefix):
-                applicable = True
-                break
-        if not applicable:
-            self.skipTest("This test only works with %s types for now"
-                          % list(supported))
-
     def setUp(self):
         super(TestAPI, self).setUp()
         if self.url is None:
@@ -75,10 +60,7 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.member_id = self._get_random_uuid()
         self._coord = tooz.coordination.get_coordinator(self.url,
                                                         self.member_id)
-        try:
-            self._coord.start()
-        except tooz.coordination.ToozConnectionError as e:
-            raise testcase.TestSkipped(str(e))
+        self._coord.start()
 
     def tearDown(self):
         self._coord.stop()
@@ -117,6 +99,31 @@ class TestAPI(testscenarios.TestWithScenarios,
         for group_id in groups_ids:
             self.assertTrue(group_id in created_groups)
 
+    def test_delete_group(self):
+        self._coord.create_group(self.group_id).get()
+        all_group_ids = self._coord.get_groups().get()
+        self.assertTrue(self.group_id in all_group_ids)
+        self._coord.delete_group(self.group_id).get()
+        all_group_ids = self._coord.get_groups().get()
+        self.assertFalse(self.group_id in all_group_ids)
+        join_group = self._coord.join_group(self.group_id)
+        self.assertRaises(tooz.coordination.GroupNotCreated,
+                          join_group.get)
+
+    def test_delete_group_non_existent(self):
+        delete = self._coord.delete_group(self.group_id)
+        self.assertRaises(tooz.coordination.GroupNotCreated,
+                          delete.get)
+
+    def test_delete_group_non_empty(self):
+        self._coord.create_group(self.group_id).get()
+        self._coord.join_group(self.group_id).get()
+        delete = self._coord.delete_group(self.group_id)
+        self.assertRaises(tooz.coordination.GroupNotEmpty,
+                          delete.get)
+        self._coord.leave_group(self.group_id)
+        self._coord.delete_group(self.group_id).get()
+
     def test_join_group(self):
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
@@ -124,8 +131,7 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.assertTrue(self.member_id in member_list)
 
     def test_join_nonexistent_group(self):
-        group_id_test = self._get_random_uuid()
-        join_group = self._coord.join_group(group_id_test)
+        join_group = self._coord.join_group(self.group_id)
         self.assertRaises(tooz.coordination.GroupNotCreated,
                           join_group.get)
 
@@ -249,7 +255,8 @@ class TestAPI(testscenarios.TestWithScenarios,
         self.assertTrue(member_id_test2 not in members_ids)
 
     def test_timeout(self):
-        self.skipIfNotSupported(self.timeout_capable)
+        if not getattr(self, "timeout_capable", False):
+            self.skipTest("This test only works with timeout capable drivers")
         member_id_test2 = self._get_random_uuid()
         client2 = tooz.coordination.get_coordinator(self.url,
                                                     member_id_test2)
@@ -260,12 +267,30 @@ class TestAPI(testscenarios.TestWithScenarios,
         members_ids = self._coord.get_members(self.group_id).get()
         self.assertTrue(self.member_id in members_ids)
         self.assertTrue(member_id_test2 in members_ids)
+
+        # Watch the group, we want to be sure that when client2 is kicked out
+        # we get an event.
+        self._coord.watch_leave_group(self.group_id, self._set_event)
+
         time.sleep(3)
         self._coord.heartbeat()
         time.sleep(3)
+
+        # Now client2 has timed out!
+
         members_ids = self._coord.get_members(self.group_id).get()
+        while True:
+            if self._coord.run_watchers():
+                break
         self.assertTrue(self.member_id in members_ids)
         self.assertTrue(member_id_test2 not in members_ids)
+        # Check that the event has been triggered
+        self.assertIsInstance(self.event,
+                              tooz.coordination.MemberLeftGroup)
+        self.assertEqual(member_id_test2,
+                         self.event.member_id)
+        self.assertEqual(self.group_id,
+                         self.event.group_id)
 
     def _set_event(self, event):
         self.event = event
@@ -405,7 +430,7 @@ class TestAPI(testscenarios.TestWithScenarios,
 
         client3 = tooz.coordination.get_coordinator(self.url, self.member_id)
         client3.start()
-        client3.leave_group(self.group_id)
+        client3.leave_group(self.group_id).get()
 
         # Only works for clients that have access to the groups they are part
         # of, to ensure that after we got booted out by client3 that this
